__all__ = [
    "is_valid_ipv4",
]


LOCALHOST = "localhost"


def is_valid_ipv4(ip: str) -> bool:
    if ip == LOCALHOST:
        return True

    locs = ip.split(".")
    if len(locs) != 4:
        return False

    try:
        for loc in map(int, locs):
            if not(0 <= loc <= 255):
                return False
    except ValueError:
        return False

    return True



from abc import ABCMeta, abstractmethod
import typing as t


__all__ = []


class FlexibleLocation(metaclass=ABCMeta):
    """Base class of flexible location.

    Location is concept representing each pieces of path of URI, i.e.
    locations can configures a path of URI by joining them with
    separator `/`.

    Location has type of `str` or subclasses of this class. If a location
    is `str`, then the location is called a static location. Otherwise,
    the location is called a flexible location. Flexible locations means
    location with some kind of logical rules of strings. The logic of rules
    can be implemented on `is_valid` method, returning if specified string
    is valid in the rules.
    """

    @abstractmethod
    def is_valid(self, loc: str) -> bool:
        pass

    def __str__(self) -> str:
        return self.__class__.__name__


Location_t = t.Union[str, FlexibleLocation]
StaticLocation_t = str
Uri_t = t.Tuple[Location_t, ...]


def is_flexible_uri(uri: Uri_t) -> bool:
    """Judge if specified `uri` has one or more flexible location.

    Args:
        uri: URI pattern to be judged.

    Returns:
        True if specified `uri` has one or more flexible location,
        False otherwise.
    """
    for loc in uri:
        if isinstance(loc, FlexibleLocation):
            return True
    return False


def is_duplicated_uri(uri_1: Uri_t, uri_2: Uri_t) -> bool:
    """Judge if a couple of specified URI patterns has same pattern.

    Args:
        uri_1: URI pattern to be judged
        uri_2: URI pattern to be judged

    Returns:
        True if two URIs has same pattern, False otherwise.
    """
    if len(uri_1) == len(uri_2):
        for loc_1, loc_2 in zip(uri_1, uri_2):
            if isinstance(loc_1, FlexibleLocation):
                continue
            if isinstance(loc_2, FlexibleLocation):
                continue

            if loc_1 != loc_2:
                break
        else:
            return True
    return False


class AsciiDigitLocation(FlexibleLocation):
    """Flexible location representing location of number in ASCII code.
    """

    def __init__(self, digits: int) -> None:
        """
        Args:
            digits: Number of digits which the location accepts.

        Raises:
            ValueError: Raised if `digits` is 0 or less.
        """
        if digits < 1:
            raise ValueError("'digits' must be bigger than 0.")

        self._digits = digits

    def is_valid(self, loc: str) -> bool:
        """Judge if specified `loc` is valid location or not.

        Args:
            loc: Location to be judged.

        Returns:
            True if specified location is valid, False otherwise.
        """
        return loc.isascii() and loc.isdigit() and len(loc) == self._digits


class AnyStringLocation(FlexibleLocation):
    """Flexible location representing string with no rules.
    """

    def __init__(self, max: t.Optional[int] = None) -> None:
        """
        Note:
            If the argument `max` is `None`, then any length of string
            will be accepted.

        Args:
            max: Max length of string of location.

        Raises:
            ValueError: Raised if `max` is 0 or less.
        """
        if max and max < 1:
            raise ValueError("'max' must be bigger than 0.")

        self._max = max

    def is_valid(self, loc: str) -> bool:
        """Judge if specified `loc` is valid location or not.

        Args:
            loc: Location to be judged.

        Returns:
            True if specified location is valid, False otherwise.
        """
        if self._max is None:
            return True

        len_loc = len(loc)
        if len_loc > self._max or len_loc == 0:
            return False
        return True



from __future__ import annotations
import dataclasses
import multiprocessing
import signal
import sys
import time
import typing as t
import wsgiref.simple_server

from bamboo.app import WSGIApp
from bamboo.util.string import ColorCode, insert_colorcode


__all__ = []


@dataclasses.dataclass
class WSGIServerForm:
    """Dataclass to register a server form.

    This dataclass would be used to register as a form
    into test classes such as WSGITestExecutor.

    Args:
        host: Hostname of the server.
        port: Port of the server.
        app: WSGIApp object with implemented Endpoints.
        path_log: Path log of the server will be written.
    """
    host: str
    port: int
    app: WSGIApp
    path_log: str


def serve_at(form: WSGIServerForm) -> None:
    """Subroutine for server application called at a child process.

    Args:
        form: Dataclass describing information of the server application.
    """
    server = wsgiref.simple_server.make_server(form.host, form.port, form.app)
    f_log = open(form.path_log, "wt")
    sys.stdout = f_log
    sys.stderr = f_log

    def server_close(signalnum, frame):
        print()
        f_log.flush()
        f_log.close()
        sys.exit()

    signal.signal(signal.SIGTERM, server_close)
    signal.signal(signal.SIGINT, server_close)
    server.serve_forever()


class WSGITestExecutor:
    """Utility class that can execute server applications at child processes.

    This class can be used in test scripts for holding several server forms
    and running the application based on the forms at the child processes.
    Also this object has feature of context manager and developers may use
    it to kill the processes safely.
    """

    def __init__(self, *forms: WSGIServerForm) -> None:
        """
        Args:
            *forms: Dataclass describing information of
                the server application.
        """
        self._forms: t.List[WSGIServerForm] = []
        self._children: t.List[multiprocessing.Process] = []

        self.add_forms(*forms)

    def add_forms(self, *forms: WSGIServerForm) -> None:
        """Add forms with information of server applications.

        Args:
            *foms: Dataclass describing information of
                the server application
        """
        for form in forms:
            self._forms.append(form)

    def start_serve(self, waiting: float = 0.05) -> WSGITestExecutor:
        """Run registered server applications at child processes.

        This object has feature of context manager and the method
        returns the object itself. So developer can use the with
        sentence and in it, can define logic of clients.

        Args:
        waiting: Waiting time after running the processes.

        Returns:
            This object itself.

        Examples:
            ```python
            >>> holder = WSGITestExecutor(form)
            >>> with holder.start_serve():
            ...    res = http_get("http://localhost:8000/image")
            >>> print(res.body)
            ```
        """
        for form in self._forms:
            child = multiprocessing.Process(target=serve_at, args=(form,))
            child.start()
            self._children.append(child)
        time.sleep(waiting)

        return self

    def close(self, pop: bool = True) -> None:
        """Kill the all child processes derived from registered forms.

        Args:
            pop: If the registered forms is to be removed.
        """
        for child in self._children:
            child.terminate()
            child.join()
            child.close()

        if pop:
            self._children.clear()

    def exec(
        self,
        func: t.Callable[[t.Tuple[t.Any, ...]], None],
        args: t.Tuple[t.Any, ...] = (),
        waiting: float = 0.1,
    ) -> None:
        """Executes a simple client-server test.

        Args:
            func: Function executed after all the server applications start.
            args: Arguments of the func.
            waiting: Waiting time after running the applications.
        """
        with self.start_serve(waiting=waiting):
            func(*args)

    @staticmethod
    def debug(
        app: WSGIApp,
        host: str = "localhost",
        port: int = 8000,
    ) -> None:
        """Executes a server application for debug.

        This method is a kind of shorcut for launching a server application
        and can be used to debug the application. If you want to deploy an
        application made with Bamboo, consider to use another WSGI server
        application for production.

        Args:
            app: WSGIApp object with implemented Endpoints.
            host: Hostname of the server.
            port: Port of the server.
        """
        server = wsgiref.simple_server.make_server(host, port, app)
        try:
            print(f"Hosting on {host}:{port} ...")
            print(insert_colorcode(
                "WARNING: This is debug mode. "
                "Do not use it in your production deployment.",
                ColorCode.RED
            ))
            server.serve_forever()
        except KeyboardInterrupt:
            server.server_close()
            print()

    def __enter__(self) -> None:
        pass

    def __exit__(self, type, value, traceback) -> None:
        self.close()



import unittest

from bamboo import (
    ContentType,
    JsonApiData,
    MediaTypes,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import data_format
from bamboo.util.time import get_datetime_rfc822

from . import PATH_IMAGE, get_log_name


NAME_SERVER = "Mocker"
PATH_SERVER_LOG_1 = get_log_name(__file__, "1")
PATH_SERVER_LOG_2 = get_log_name(__file__, "2")
PATH_SERVER_LOG_3 = get_log_name(__file__, "3")


class InfoResponse(JsonApiData):

    server_name: str
    current_time: str


class MockInfoEndpoint(WSGIEndpoint):

    @data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        body = {
            "server_name": NAME_SERVER,
            "current_time": get_datetime_rfc822()
        }
        self.send_json(body)


class MockImageEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        with open(PATH_IMAGE, "rb") as f:
            image = f.read()

        self.send_body(image, content_type=ContentType(MediaTypes.jpeg))


class TestHTTPRequest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.urls_info = [f"http://localhost:800{i}/mock/info" for i in range(3)]
        cls.urls_image = [f"http://localhost:800{i}/mock/image" for i in range(3)]

        app_1 = WSGIApp()
        app_2 = WSGIApp()
        app_3 = WSGIApp()

        for app in (app_1, app_2, app_3):
            app.route("mock", "info")(MockInfoEndpoint)
            app.route("mock", "image")(MockImageEndpoint)

        form_1 = WSGIServerForm("", 8000, app_1, PATH_SERVER_LOG_1)
        form_2 = WSGIServerForm("", 8001, app_2, PATH_SERVER_LOG_2)
        form_3 = WSGIServerForm("", 8002, app_3, PATH_SERVER_LOG_3)
        cls.executor = WSGITestExecutor(form_1, form_2, form_3).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_get_info(self):
        for url in self.urls_info:
            res = http.get(url, datacls=InfoResponse)
            data = res.attach()

            self.assertTrue(isinstance(data, InfoResponse))
            self.assertEqual(data.server_name, "Mocker")

    def test_get_image(self):
        with open(PATH_IMAGE, "rb") as f:
            image_ideal = f.read()

        for url in self.urls_image:
            res = http.get(url)
            data = res.body
            self.assertEqual(image_ideal, data)


if __name__ == "__main__":
    unittest.main()



import unittest

from bamboo import (
    ASGIHTTPApp,
    ASGIHTTPEndpoint,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import basic_auth
from bamboo.util.convert import encode_base64_string
from bamboo.util.string import rand_string

from . import get_log_name
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app_asgi = ASGIHTTPApp()
app_wsgi = WSGIApp()
PATH_ASGI_SERVER_LOG = get_log_name(__file__, "asgi")
PATH_WSGI_SERVER_LOG = get_log_name(__file__, "wsgi")
RANDOM_CREDENTIALS = [(rand_string(10), rand_string(10)) for _ in range(10)]


def make_basic_credential(user_id: str, pw: str) -> str:
    return encode_base64_string(f"{user_id}:{pw}")


@app_asgi.route()
class TestASGIHTTPEndpoint(ASGIHTTPEndpoint):

    @basic_auth()
    async def do_HEAD(self, user_id: str, pw: str) -> None:
        self.send_only_status()


@app_wsgi.route()
class TestWSGIEndpoint(WSGIEndpoint):

    @basic_auth()
    def do_HEAD(self, user_id: str, pw: str) -> None:
        self.send_only_status()


class TestStickyRestrictsClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form_asgi = ASGIHTTPServerForm("", 8000, app_asgi, PATH_ASGI_SERVER_LOG)
        form_wsgi = WSGIServerForm("", 8001, app_wsgi, PATH_WSGI_SERVER_LOG)
        cls.executor_asgi = ASGIHTTPTestExecutor(form_asgi).start_serve()
        cls.executor_wsgi = WSGITestExecutor(form_wsgi).start_serve()
        cls.uri_asgi = "http://localhost:8000"
        cls.uri_wsgi = "http://localhost:8001"

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor_asgi.close()
        cls.executor_wsgi.close()

    def test_asgi(self):
        for user_id, pw in RANDOM_CREDENTIALS:
            formatted = make_basic_credential(user_id, pw)
            headers = {"Authorization": "Basic " + formatted}
            with http.head(self.uri_asgi, headers=headers) as res:
                self.assertTrue(res.ok)

    def test_wsgi(self):
        for user_id, pw in RANDOM_CREDENTIALS:
            formatted = make_basic_credential(user_id, pw)
            headers = {"Authorization": "Basic " + formatted}
            with http.head(self.uri_wsgi, headers=headers) as res:
                self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()



import http.client as client
import os
import unittest

from bamboo import (
    ContentType,
    JsonApiData,
    MediaTypes,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import Response
from bamboo.sticky.http import data_format
from bamboo.util.time import get_datetime_rfc822

from . import PATH_IMAGE, get_log_name


app = WSGIApp()
NAME_SERVER = "Mocker"
PATH_SERVER_LOG = get_log_name(__file__)


class InfoResponse(JsonApiData):

    server_name: str
    current_time: str


@app.route("mock", "info")
class MockInfoEndpoint(WSGIEndpoint):

    @data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        body = {
            "server_name": NAME_SERVER,
            "current_time": get_datetime_rfc822()
        }
        self.send_json(body)


@app.route("mock", "image")
class MockImageEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        with open(PATH_IMAGE, "rb") as f:
            image = f.read()

        self.send_body(image, content_type=ContentType(MediaTypes.jpeg))


class TestResponse(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.conn = client.HTTPConnection("localhost", 8000)
        form = WSGIServerForm("", 8000, app, PATH_SERVER_LOG)
        cls.executor = WSGITestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_body(self):
        with open(PATH_IMAGE, "rb") as f:
            image_ideal = f.read()

        uri = "/" + os.path.join("mock", "image")
        self.conn.request("GET", uri)
        _res = self.conn.getresponse()
        with Response(self.conn, _res) as res:
            self.assertEqual(image_ideal, res.body)

    def test_attach_datacls(self):
        uri = "/" + os.path.join("mock", "info")
        self.conn.request("GET", uri)
        _res = self.conn.getresponse()
        with Response(self.conn, _res, datacls=InfoResponse) as res:
            data = res.attach()

        self.assertTrue(isinstance(data, InfoResponse))
        self.assertEqual(data.server_name, NAME_SERVER)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import typing as t

from ..api import ApiData


__all__ = [
    "Response",
    "Schemes",
]


ResponseData_t = t.TypeVar("ResponseData_t", bound=ApiData)


class _Schemes:

    HTTP = "http"
    HTTPS = "https"

    _schemes = set((HTTP, HTTPS))
    __instance = None

    def __new__(cls) -> _Schemes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self):
        return iter(self._schemes)

    def __contains__(self, item: str):
        return item in self._schemes


Schemes = _Schemes()


from bamboo.request.response import Response



import random
import unittest

from bamboo import (
    AnyStringLocation,
    AsciiDigitLocation,
    JsonApiData,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import data_format
from bamboo.util.string import rand_string
from bamboo.util.time import get_datetime_rfc822

from . import get_log_name


app = WSGIApp()
PATH_SERVER_LOG = get_log_name(__file__)


class InfoResponse(JsonApiData):

    name: str
    datetime: str


@app.route(AnyStringLocation(), "info")
class InfoEndpoint(WSGIEndpoint):

    @data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        name, = self.flexible_locs
        body = {"name": name, "datetime": get_datetime_rfc822()}
        self.send_json(body)


class CalculationResult(JsonApiData):

    result: int


@app.route(AsciiDigitLocation(10), AsciiDigitLocation(10), "multiply")
class MultiplyEndpoint(WSGIEndpoint):

    @data_format(input=None, output=CalculationResult)
    def do_GET(self) -> None:
        num_1, num_2 = self.flexible_locs
        body = {"result": int(num_1) * int(num_2)}
        self.send_json(body)


class FlexibleLocsTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form = WSGIServerForm("", 8000, app, PATH_SERVER_LOG)
        cls.executor = WSGITestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_info(self):
        names = [rand_string(10) for _ in range(100)]

        for name in names:
            uri = f"http://localhost:8000/{name}/info"
            with http.get(uri, datacls=InfoResponse) as res:
                if res.ok:
                    body = res.attach()
                    self.assertEqual(body.name, name)
                else:
                    print(f"Request failed. Status code: {res.status}")

    def test_multiply(self):
        get_num = lambda: random.randint(10**9, 10**10 - 1)
        num_pairs = [(get_num(), get_num()) for _ in range(100)]

        for num_1, num_2 in num_pairs:
            uri = f"http://localhost:8000/{num_1}/{num_2}/multiply"
            with http.get(uri, datacls=CalculationResult) as res:
                if res.ok:
                    body = res.attach()
                    self.assertEqual(body.result, num_1 * num_2)
                else:
                    print(f"Request failed. Status code: {res.status}")


if __name__ == "__main__":
    unittest.main()



import unittest

from bamboo import (
    ASGIHTTPEndpoint,
    ErrInfo,
    WSGIEndpoint,
)
from bamboo.sticky.http import HTTPErrorConfig, may_occur


ERRORS = [type(f"TestErr{i}", (ErrInfo,), {}) for i in range(10)]


class TestWSGIEndpoint(WSGIEndpoint):

    @may_occur(*ERRORS)
    def do_GET(self) -> None:
        pass


class TestASGIHTTPEndpoint(ASGIHTTPEndpoint):

    @may_occur(*ERRORS)
    async def do_GET(self) -> None:
        pass


class TestSticyMayOccur(unittest.TestCase):

    def check_registered(self, callback):
        config = HTTPErrorConfig(callback)
        errors_registered = config.get()

        for err in ERRORS:
            self.assertIn(err, errors_registered)

    def test_wsgi(self):
        self.check_registered(TestWSGIEndpoint.do_GET)

    def test_asgi(self):
        self.check_registered(TestASGIHTTPEndpoint.do_GET)


if __name__ == "__main__":
    unittest.main()



__version__ = "0.7.11"


from bamboo.api import (
    ApiData,
    BinaryApiData,
    InvalidAnnotationError,
    JsonApiDataBuilder,
    JsonApiData,
    ApiValidationFailedError,
    XWWWFormUrlEncodedDataBuilder,
    XWWWFormUrlEncodedData,
)
from bamboo.app import (
    AppBase,
    ASGIHTTPApp,
    ParcelConfig,
    VersionConfig,
    WSGIApp,
)
from bamboo.base import (
    ASGIHTTPEvents,
    ASGIWebSocketEvents,
    AuthSchemes,
    ContentType,
    ContentTypeHolder,
    DEFAULT_CONTENT_TYPE_JSON,
    DEFAULT_CONTENT_TYPE_PLAIN,
    HTTPMethods,
    HTTPStatus,
    MediaTypes,
)
from bamboo.endpoint import (
    ASGIEndpointBase,
    ASGIHTTPEndpoint,
    EndpointBase,
    HTTPMixIn,
    StatusCodeAlreadySetError,
    WSGIEndpoint,
    WSGIEndpointBase,
)
from bamboo.error import (
    ApiErrInfo,
    DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_HEADER_NOT_FOUND_ERROR,
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    DEFAULT_NOT_APPLICABLE_IP_ERROR,
    DEFAULT_NOT_FOUND_ERROR,
    DEFAULT_QUERY_PARAM_NOT_FOUND_ERROR,
    ErrInfo,
    get_auth_realm,
    get_default_auth_realm,
)
from bamboo.io import (
    BufferedBinaryIterator,
    BufferedConcatIterator,
    BufferedFileIterator,
    BufferedIteratorWrapper,
    BufferedStreamIterator,
    ResponseBodyIteratorBase,
)
from bamboo.location import (
    AnyStringLocation,
    AsciiDigitLocation,
    FlexibleLocation,
    is_duplicated_uri,
    is_flexible_uri,
)
from bamboo.router import (
    DuplicatedUriRegisteredError,
    Router
)
from bamboo.test import (
    WSGIServerForm,
    WSGITestExecutor
)



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import json
import typing as t
from urllib.parse import parse_qs

from .base import (
    ContentType,
    ContentTypeHolder,
    DEFAULT_CONTENT_TYPE_JSON,
    DEFAULT_CONTENT_TYPE_PLAIN,
    MediaTypes,
)
from .util.deco import cached_property, class_property
from .util.typing import get_args, get_origin


__all__ = []


class ApiData(ContentTypeHolder, metaclass=ABCMeta):
    """Base class to describe input/output data format on Endpoints as APIs.

    Subclasses of this class can be used an argument of `data_format`
    decorator for callbacks on Endpoints. If input/output parameters of
    the `data_format` are specified, the decorator validates if raw data,
    typically a `bytes` object, has expected data format.

    Subclasses of this class can define their own data formats in `__init__`
    methods. Developers should implement `__init__` methods of the subclasses
    such that each objects has data with expected formats by implementors.

    Note:
        This class is an abstract class. So, don't initilize it and specify it
        as an argument of the `data_format` decorator.

        Subclasses of this class should validate if raw data given to
        `__init__` methods has expected formats. If the validation failes,
        the classes MUST raise `ValidataionFailedError` to announce the
        failure to the `data_format` decorator.
    """

    @classmethod
    @abstractmethod
    def __validate__(cls, raw: bytes, content_type: ContentType) -> ApiData:
        """
        Args:
            raw : Raw data to be validated.
            content_type : Values of `Content-Type` header.
        """
        pass


class ApiValidationFailedError(Exception):
    """Raised if type validation of data failes."""
    pass


class BinaryApiData(ApiData):
    """API data with no format.

    This class can be used to describe raw data with no data format. So,
    any received data from clients is acceptable on the class.

    Examples:
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=BinaryApiData, output=None)
            def do_GET(self, rec_body: BinaryApiData) -> None:
                # get raw data of request body
                raw_data = rec_body.raw
        ```
    """

    def __init__(self, data: bytes) -> None:
        """
        Args:
            data: Binary data.
        """
        self._data = data

    @classmethod
    def __validate__(cls, raw: bytes, content_type: ContentType) -> BinaryApiData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.

        Returns:
            The object of the class validated, if validation successes.

        Note:
            In objects of this class, `content_type` is not used even if any
            `content_type` is specified.
        """
        if not isinstance(raw, bytes):
            raise ApiValidationFailedError(f"'raw' must be a 'bytes'.")
        return cls(raw)

    @property
    def raw(self) -> bytes:
        """Raw data of input binary."""
        return self._data

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `text/plain`.
        """
        return DEFAULT_CONTENT_TYPE_PLAIN


class InvalidAnnotationError(Exception):
    """Raised if a type not convertable to JSON is included of type hints."""
    pass


ListJsonable_t = t.TypeVar("ListJsonable_t")


class JsonApiDataBuilder:
    """Builder for `JsonApiData`.

    This utility class has some class methods for validating JSON data and
    building `JsonApiData` objects from raw data.

    Attributes:
        NoneType (type): The type of None.
        TYPES_ARGS (Tuple[type]): Tuple of types acceptable to `JsonApiData`
            with no arguments.
        TYPES_ARGS_SET (Set[type]): Set of types acceptable to `JsonApiData`
            with no arguments.
        TYPES_ORIGIN (Tuple[type]): Tuple of types acceptable to
            `JsonApiData` with arguments.
    """

    NoneType = type(None)
    TYPES_ARGS = (int, float, str, bool, dict, list, NoneType)
    TYPES_ARGS_SET = set(TYPES_ARGS)
    TYPES_ORIGIN = (list, t.Union)

    @classmethod
    def check_annotations(cls, *types: t.Type) -> None:
        """Check if types are convertable to JSON data.

        Args:
            *types: Classes to be checked.

        Raises:
            InvalidAnnotationError: Raised if a type not convertable to
                JSON data is detected.
        """
        def checker(types: t.Sequence[t.Type]) -> None:
            for type_ in types:
                origin = get_origin(type_)
                if origin is None:
                    if issubclass(type_, cls.TYPES_ARGS):
                        continue
                    elif issubclass(type_, JsonApiData):
                        checker(t.get_type_hints(JsonApiData).values())
                    else:
                        raise InvalidAnnotationError(
                            f"{type_.__name__} is an unacceptable as a type "
                            "of attributes of JsonApiData."
                        )
                else:
                    args = get_args(type_)
                    if origin == list:
                        if not len(args):
                            raise InvalidAnnotationError(
                                "List with no arguments is not acceptable. "
                                "Specify an argument of type to list."
                            )
                        checker([args[0]])
                    elif origin == t.Union:
                        # Only Optional is admitted.
                        if not (len(args) == 2 and args[1] == cls.NoneType):
                            raise InvalidAnnotationError(
                                "Attritbutes with serveral types are not "
                                "acceptable. Only 'Optional' can be used as "
                                "type hints."
                            )

                        checker([args[0]])
                    else:
                        raise InvalidAnnotationError(
                            f"{type_.__name__} is not acceptable as a type of"
                            " attributes of JsonApiData."
                        )
        checker(types)

    @classmethod
    def has_valid_annotations(cls, apiclass: t.Type[JsonApiData]) -> None:
        """Check if `JsonApiData` has only arguments convertable to JSON.

        Args:
            apiclass: A subclass of the `JsonApiData`.

        Raises:
            InvalidAnnotationError: Raised if a type not convertable to
                JSON data is detected.
        """
        cls.check_annotations(*tuple(t.get_type_hints(apiclass).values()))

    @classmethod
    def validate_obj(cls, obj: t.Any, objtype: t.Type) -> bool:
        """Validate if given object is acceptable as a value of `JsonApiData`.

        Args:
            obj: Value to be validated.
            objtype: Type to be validated.

        Returns:
            True if the obj and objtype is acceptable to `JsonApiData`,
            False otherwise.
        """
        origin = get_origin(objtype)
        if origin is None:
            if objtype in cls.TYPES_ARGS_SET:
                return isinstance(obj, objtype)
            return issubclass(objtype, JsonApiData) and isinstance(obj, dict)

        if origin == list:
            if len(obj):
                item = obj[0]
                inner_type = get_args(objtype)[0]
                return cls.validate_obj(item, inner_type)
            return True
        elif origin == t.Union:
            is_good = False
            for arg in get_args(objtype):
                is_good = is_good or cls.validate_obj(obj, arg)
            return is_good
        else:
            return False

    @classmethod
    def build(
        cls,
        apiclass: t.Type[JsonApiData],
        data: t.Dict[str, t.Any]
    ) -> JsonApiData:
        """Build new `JsonApiData` object from raw data.

        Args:
            apiclass: Class of new object to be generated.
            data: Raw data converted as a dictionary.

        Returns:
            New object built.

        Raises:
            ApiValidationFailedError: Raised if type validation of raw
                data failes.
        """
        instance = apiclass.__new__(apiclass)
        annotations = t.get_type_hints(apiclass)

        # NOTE
        #   Ignore keys of data which is not defined in the apiclass.
        for key_def, type_def in annotations.items():
            if not (key_def in data or hasattr(apiclass, key_def)):
                raise ApiValidationFailedError(
                    f"Raw data has no key of '{key_def}'."
                )

            val = data.get(key_def)
            if not cls.validate_obj(val, type_def):
                raise ApiValidationFailedError(
                    "Invalid type was detected in received json data. "
                    f"Expected: {type_def.__name__}; "
                    f"Detected: {val.__class__.__name__}"
                )

            # Set values of attributes to an instance of the apiclass.
            origin = get_origin(type_def)
            if origin is None:
                if issubclass(type_def, JsonApiData):
                    setattr(instance, key_def, cls.build(type_def, val))
                else:
                    setattr(instance, key_def, val)
            else:
                if origin == list:
                    setattr(instance, key_def, cls._build_list(type_def, val))
                elif origin == t.Union:
                    possible = get_args(type_def)[0]
                    if possible in cls.TYPES_ARGS_SET:
                        setattr(instance, key_def, val)
                    elif possible == list:
                        setattr(
                            instance,
                            key_def,
                            cls._build_list(possible, val)
                        )
                    elif issubclass(possible, JsonApiData):
                        setattr(instance, key_def, cls.build(type_def, val))
                    else:
                        raise ApiValidationFailedError(
                            f"{possible.__name__} is not acceptable as a type"
                            " of attributes of JsonApiData."
                        )
        return instance

    @classmethod
    def _build_list(
        cls,
        type_def: ListJsonable_t,
        val: list
    ) -> ListJsonable_t:
        """Build list convertable to JSON data.

        Args:
            type_def: Type of object to be generated, must be list with
                an argument.
            val: Raw data converted as a list.

        Returns:
            The generated object.

        Raises:
            ValueError: Raised if `type_def` is not list with an argument.
            ApiValidationFailedError: Raised if the argument of `type_def` is
                not acceptable.
        """
        if not len(val):
            return val
        if not (get_origin(type_def) == list and len(get_args(type_def))):
            raise ValueError("'type_def' must be list with an argument.")

        # Check type of inner value
        inner_type = get_args(type_def)[0]
        if inner_type in cls.TYPES_ARGS_SET:
            return val
        elif issubclass(inner_type, JsonApiData):
            return [cls.build(inner_type, item) for item in val]
        elif get_origin(inner_type) == list:
            return [cls._build_list(inner_type, item) for item in val]
        else:
            raise ApiValidationFailedError(
                f"{inner_type.__name__} is not acceptable as a type of "
                "attributes of JsonApiData."
            )


class JsonApiData(ApiData):
    """API data with JSON format.

    This class can be used to describe data with JSON format. This class
    should be inheritted and its several class-attributes should be
    defined in the subclass. Developers must define type hints of the
    class-attribtues, which are used to validate if raw data has format
    the type hints define.

    Examples:
        - Defining subclass of this class
        ```python
        class User(JsonApiData):
            name: str
            email: str
            age: int

        class MockApiData(JsonApiData):
            users: List[User]
        ```

        - Validating received data
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=MockApiData, output=None)
            def do_GET(self, rec_body: MockApiData) -> None:
                # Do something...

                # Example
                for user in rec_body.users:
                    print(f"user name : {user.name}")
        ```
    """

    def __init_subclass__(cls) -> None:
        JsonApiDataBuilder.has_valid_annotations(cls)

    # NOTE
    #   DO NOT override the method. This class should be used only
    #   by defining attributes and their types.
    def __init__(self, **data: t.Any) -> None:
        self._dict = data

        mapped = JsonApiDataBuilder.build(self.__class__, data)
        self.__dict__.update(mapped.__dict__)

    @classmethod
    def __validate__(cls, raw: bytes, content_type: ContentType) -> ApiData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.
        """
        super().__init__(raw, content_type)

        if content_type.charset is None:
            content_type.charset = "UTF-8"

        if not cls.verify_content_type(content_type):
            raise ApiValidationFailedError(
                "Media type of 'Content-Type' header was not "
                f"{MediaTypes.json}, but {content_type.media_type}."
            )

        try:
            raw = raw.decode(encoding=content_type.charset)
            data = json.loads(raw)
        except UnicodeDecodeError:
            raise ApiValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{content_type.charset}, but not corresponded."
            )
        except json.decoder.JSONDecodeError:
            raise ApiValidationFailedError(
                "Decoding raw data failed."
                "The raw data had invalid JSON format."
            )
        return cls(**data)

    @property
    def dict(self) -> t.Dict[str, t.Any]:
        return self._dict

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `application/json; charset=UTF-8`.
        """
        return DEFAULT_CONTENT_TYPE_JSON


class XWWWFormUrlEncodedDataBuilder:
    """Builder for `XWWWFormUrlEncodedData`.

    This utility class has some class methods for validating
    request body with the `x-www-urlencoded` format.

    Attributes:
        TYPES_ARGS (Tuple[type]): Tuple of types acceptable to
            `XWWWFormUrlEncodedData`.
    """

    TYPES_ARGS = (str,)

    @classmethod
    def check_annotations(cls, *types: t.Type) -> None:
        """Check if types are acceptable to `XWWWFormUriEncodedData` class.

        Args:
            *types: Classes to be checked.

        Raises:
            InvalidAnnotationError: Raised if a type not acceptable.
        """
        for type_ in types:
            if type_ not in cls.TYPES_ARGS:
                raise InvalidAnnotationError(
                    f"{type_.__name__} is unacceptable as a type "
                    "of attributes of XWWWFormUrlEncodedData."
                )

    @classmethod
    def has_valid_annotations(
        cls,
        apiclass: t.Type[XWWWFormUrlEncodedData]
    ) -> None:
        """Check if `XWWWFormUrlEncodedData` has only arguments with
        acceptable types.

        Args:
            apiclass: A subclass of the `XWWWFormUrlEncodedData`.

        Raises:
            InvalidAnnotationError: Raised if a type not acceptable.
        """
        cls.check_annotations(*tuple(t.get_type_hints(apiclass).values()))

    @classmethod
    def build_dict(
        cls,
        apiclass: t.Type[XWWWFormUrlEncodedData],
        **data: str,
    ) -> XWWWFormUrlEncodedData:
        instance = apiclass.__new__(apiclass)
        annotations = t.get_type_hints(apiclass)

        for key in annotations.keys():
            if key not in data:
                raise ApiValidationFailedError(
                    f"Raw data has no key of '{key}'."
                )
            if type(data[key]) not in cls.TYPES_ARGS:
                raise ApiValidationFailedError(
                    "Invalid type was detected. "
                    f"Expected {apiclass.__name__}.{key} was a str."
                )
            setattr(instance, key, data[key])
        return instance

    @classmethod
    def build(
        cls,
        apiclass: t.Type[XWWWFormUrlEncodedData],
        data: str
    ) -> XWWWFormUrlEncodedData:
        """Build new `XWWWFormUrlEncodedData` object from raw data.

        Args:
            apiclass: Class of new object to be generated.
            data: Raw data already decoded.

        Returns:
            New object built.

        Raises:
            ApiValidationFailedError: Raised if type validation of
                raw data failes.
        """
        parsing = {}
        for key, val in parse_qs(data).items():
            if len(val) > 1:
                raise ApiValidationFailedError(
                    f"Duplicated keys '{key}' were detected."
                )
            parsing[key] = val[0]
        return cls.build_dict(apiclass, **parsing)


class XWWWFormUrlEncodedData(ApiData):
    """API data with `x-www-form-urlencoded`

    This class can be used to describe data with `x-www-form-urlencoded`
    format. This class should be inheritted and its several
    class-attributes should be defined in the subclass. Developer must
    define type hints of the class-attributes, which are used to validate if
    raw data has format the type hints define. In this class, the type hints
    must be only `str`. Otherwise, `InvalidAnnotationError` will be raised.

    Examples:
        - Defining subclass of this class
        ```python
        class UserCredentials(XWWWFormUrlEncodedData):
            user_id: str
            password: str
        ```

        - Validating received data
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=UserCredentials, output=None)
            def do_GET(self, rec_body: UserCredentials) -> None:
                # Do something...

                # Example
                authenticate(rec_body.user_id, rec_body.password)
        ```
    """

    def __init_subclass__(cls) -> None:
        XWWWFormUrlEncodedDataBuilder.has_valid_annotations(cls)

    def __init__(self, **data: str) -> None:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.
        """
        mapped = XWWWFormUrlEncodedDataBuilder.build_dict(self.__class__, **data)
        self.__dict__.update(mapped.__dict__)

    @classmethod
    def __validate__(
        cls,
        raw: bytes,
        content_type: ContentType,
    ) -> XWWWFormUrlEncodedData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.

        Returns:

        """
        if content_type.charset is None:
            content_type.charset = "UTF-8"

        if not cls.verify_content_type(content_type):
            raise ApiValidationFailedError(
                "Media type of 'Content-Type' header is not "
                f"{MediaTypes.x_www_form_urlencoded}, "
                f"but {content_type.media_type}."
            )

        try:
            raw = raw.decode(encoding=content_type.charset)
        except UnicodeDecodeError:
            raise ApiValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{content_type.charset}, but not corresponded."
            )
        return XWWWFormUrlEncodedDataBuilder.build(cls, raw)

    @cached_property
    def dict(self) -> t.Dict[str, t.Any]:
        return {key: getattr(self, key) for key in t.get_type_hints(self).keys()}

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `application/x-www-form-urlencoded`.
        """
        return ContentType(MediaTypes.x_www_form_urlencoded)
